"""Connect to a Bayernluefter."""

import aiohttp
import re
import parse
from http import HTTPStatus

from typing import Any, Dict

from .convert import CONVERSION_DICT

ENDPOINT_EXPORT = "?export=1"
ENDPOINT_TEMPLATE = "/export.txt"
ENDPOINT_POWER_ON = "?power=on"
ENDPOINT_POWER_OFF = "?power=off"
ENDPOINT_BUTTON_POWER = "?button=power"
ENDPOINT_BUTTON_TIMER = "?button=timer"
ENDPOINT_SPEED = "?speed={}"


def repl_to_parse(m: re.Match):
    # prepend a v s.t. no variable begins with an underscore
    return "{{v{}}}".format(m.group(1))


def construct_url(ip_address: str) -> str:
    """Construct the URL with a given IP address."""
    if "http://" not in ip_address and "https://" not in ip_address:
        ip_address = "{}{}".format("http://", ip_address)
    ip_address = ip_address.rstrip("/")
    return ip_address


class Bayernluefter:
    """Interface to communicate with the Bayernluefter."""

    def __init__(self, ip, session: aiohttp.ClientSession) -> None:
        """Initialize the the printer."""
        self.url = construct_url(ip)
        self._session = session
        self.data = {}  # type: Dict[str, Any]
        self.data_converted = {}  # type: Dict[str, Any]
        self.template = None

    async def fetch_template(self):
        """
        Fetches the template for the export values from the Bayernluefter
        """

        bl_template = await self._request_bl(ENDPOINT_TEMPLATE)
        self.template = re.sub(r"~(.+)~", repl_to_parse, bl_template)

    async def update(self) -> None:
        """
        Retrieve the data from the printer.
        Throws ValueError if host does not support SyncThru
        """
        if self.template is None:
            await self.fetch_template()

        state = await  self._request_bl(ENDPOINT_EXPORT)
        parse_dict = parse.parse(self.template, state).named
        self.data = {
            key[1:]: value
            for key, value in parse_dict.items()
        }
        self.data_converted = {
            key: CONVERSION_DICT.get(key, str)(value) for key, value in self.data.items()
        }

    async def _request_bl(self, target):
        url = "{}{}".format(self.url, target)
        async with self._session.get(url) as response:
            if response.status != HTTPStatus.OK:
                raise ValueError("Server does not support Bayernluefter protocol.")
            return await response.text(encoding="ascii", errors="ignore")

    def raw(self) -> Dict:
        """Return all details of the Bayernluefter."""
        return self.data

    def raw_converted(self) -> Dict:
        """Return all details of the Bayernluefter, converted"""
        return self.data_converted

    async def power_on(self):
        await self._request_bl(ENDPOINT_POWER_ON)

    async def power_off(self):
        await self._request_bl(ENDPOINT_POWER_OFF)

    async def power_toggle(self):
        await self._request_bl(ENDPOINT_BUTTON_POWER)

    async def timer_toggle(self):
        await self._request_bl(ENDPOINT_BUTTON_TIMER)

    async def reset_speed(self):
        await self._request_bl(ENDPOINT_SPEED.format(0))

    async def set_speed(self, level:int):
        assert 1 <= level <= 10, "Level must be between 1 and 10"
        await self._request_bl(ENDPOINT_SPEED.format(level))



import os
from http.server import SimpleHTTPRequestHandler, HTTPServer
try:
    from http import HTTPStatus
except ImportError:
    # Backwards compatability
    import http.client as HTTPStatus
from pathlib import Path

SERVER_DIR = Path(__file__).parent or Path('.')
REVISIONS = ["rev1", "rev2"]


class BayernLuftServer(HTTPServer):

    blocked = False

    def __init__(self, revision: str, *args, **kwargs):
        assert revision in REVISIONS, f"Revision should be one of {REVISIONS}"
        self.revision = revision
        super(BayernLuftServer, self).__init__(*args, **kwargs)

    def set_blocked(self):
        self.blocked = True

    def unset_blocked(self):
        self.blocked = False


class BayernLuftHandler(SimpleHTTPRequestHandler):

    def do_GET(self):
        if self.server.blocked:
            self.send_error(403, "Access denied because server blocked")
        else:
            super(BayernLuftHandler, self).do_GET()

    def translate_path(self, path:str):
        """
        translate paths for the two files we provide
        """
        path = path.lstrip('/')
        server_path = str(SERVER_DIR.joinpath(self.server.revision).absolute())
        if path.endswith("?export=1") and (path.startswith("?") or path.startswith("index.html")):
            return os.path.join(server_path, "?export=1")
        elif path == "export.txt":
            return os.path.join(server_path, "export.txt")
        return os.path.join(server_path, "nothingtobefound")

    def send_error(self, code, message=None, explain=None):
        """
        Send syncthru error page
        :param code:
        :param message:
        :param explain:
        :return:
        """
        self.log_error("code %d, message %s", code, message)
        self.send_response(code, message)
        self.send_header('Connection', 'close')

        # Message body is omitted for cases described in:
        #  - RFC7230: 3.3. 1xx, 204(No Content), 304(Not Modified)
        #  - RFC7231: 6.3.6. 205(Reset Content)
        body = None
        if (code >= 200 and
                code not in (HTTPStatus.NO_CONTENT, HTTPStatus.RESET_CONTENT,
                             HTTPStatus.NOT_MODIFIED)):
            # HTML encode to prevent Cross Site Scripting attacks
            # (see bug #1100201)
            # Specialized error method for fronius
            with SERVER_DIR.joinpath(self.server.revision).joinpath(".error.html").open('rb') as file:
                body = file.read()
            self.send_header("Content-Type", self.error_content_type)
            self.send_header('Content-Length', str(len(body)))
        self.end_headers()

        if self.command != 'HEAD' and body:
            self.wfile.write(body)



from datetime import datetime
from enum import Enum


class SystemMode(Enum):
    Zeitgesteuert = "Zeitgesteuert"
    Kellermode = "Kellermode"
    Behaglichkeitsmode = "Behaglichkeitsmode"


def comma_float(x:str):
    if x == "N/A":
        return None
    return float(x.replace(",", "."))


def custom_bool(x:str):
    return x == "1"


CONVERSION_DICT = {
    "Date": lambda x: datetime.strptime(x, "%d.%m.%Y").date(),
    "Time": lambda x: datetime.strptime(x, "%H:%M:%S").time(),
    "MAC": lambda x: hex(int(x, 16)),
    "RSSI": int,
    "SystemMode": lambda x: SystemMode(x),
    "Speed_In": int,
    "Speed_Out": int,
    "Speed_AntiFreeze": int,
    "Temp_In": comma_float,
    "Temp_Out": comma_float,
    "Temp_Fresh": comma_float,
    "rel_Humidity_In": comma_float,
    "rel_Humidity_Out": comma_float,
    "abs_Humidity_In": comma_float,
    "abs_Humidity_Out": comma_float,
    "Efficiency": comma_float,
    "Humidity_Transport": int,
    "_SystemOn": custom_bool,
    "_FrostschutzAktiv": custom_bool,
    "_Frozen": custom_bool,
    "_AbtauMode": custom_bool,
    "_VermieterMode": custom_bool,
    "_QuerlueftungAktiv": custom_bool,
    "_MaxMode": custom_bool,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# general requirements
import unittest
from .test_structure.server_control import Server
from .test_structure.bayernluefter_mock_server import BayernLuftHandler, BayernLuftServer

# For the server in this case
import time

# For the tests
import aiohttp
import asyncio
from pyernluefter import Bayernluefter
from .web_raw.web_state import RAW_V1, PROCESSED_V1, RAW_V2, PROCESSED_V2

ADDRESS = 'localhost'


class BayernluefterWebTestRev1(unittest.TestCase):

    server = None
    server_control = None
    port = 0
    url = 'http://localhost:80'
    bayernluefter = None

    def setUp(self):
        # Create an arbitrary subclass of TCP Server as the server to be started
        # Here, it is an Simple HTTP file serving server
        handler = BayernLuftHandler

        max_retries = 10
        r = 0
        while not self.server:
            try:
                # Connect to any open port
                self.server = BayernLuftServer("rev1", (ADDRESS, 0), handler)
            except OSError:
                if r < max_retries:
                    r += 1
                else:
                    raise
                time.sleep(1)

        self.server_control = Server(self.server)
        self.port = self.server_control.get_port()
        self.url = "{}:{}".format(ADDRESS, self.port)
        # Start test server before running any tests
        self.server_control.start_server()

        async def fetch():
            async with aiohttp.ClientSession() as session:
                self.bayernluefter = Bayernluefter(self.url, session)
                await self.bayernluefter.update()

        loop = asyncio.get_event_loop()
        loop.run_until_complete(fetch())

    def test_raw(self):
        self.assertEqual(self.bayernluefter.raw(), RAW_V1)

    def test_raw_converted(self):
        self.assertEqual(self.bayernluefter.raw_converted(), PROCESSED_V1)

    def tearDown(self):
        self.server_control.stop_server()
        pass

class BayernluefterWebTestRev2(unittest.TestCase):

    server = None
    server_control = None
    port = 0
    url = 'http://localhost:80'
    bayernluefter = None

    def setUp(self):
        # Create an arbitrary subclass of TCP Server as the server to be started
        # Here, it is an Simple HTTP file serving server
        handler = BayernLuftHandler

        max_retries = 10
        r = 0
        while not self.server:
            try:
                # Connect to any open port
                self.server = BayernLuftServer("rev2", (ADDRESS, 0), handler)
            except OSError:
                if r < max_retries:
                    r += 1
                else:
                    raise
                time.sleep(1)

        self.server_control = Server(self.server)
        self.port = self.server_control.get_port()
        self.url = "{}:{}".format(ADDRESS, self.port)
        # Start test server before running any tests
        self.server_control.start_server()

        async def fetch():
            async with aiohttp.ClientSession() as session:
                self.bayernluefter = Bayernluefter(self.url, session)
                await self.bayernluefter.update()

        loop = asyncio.get_event_loop()
        loop.run_until_complete(fetch())

    def test_raw(self):
        self.assertEqual(self.bayernluefter.raw(), RAW_V2)

    def test_raw_converted(self):
        self.assertEqual(self.bayernluefter.raw_converted(), PROCESSED_V2)

    def tearDown(self):
        self.server_control.stop_server()
        pass


class NonSyncthruWebTest(unittest.TestCase):

    server = None
    server_control = None
    port = 0
    url = 'http://localhost:80'
    syncthru = None

    def test_no_syncthru(self):
        """Test that an error is thrown when no syncthru is supported"""
        # Create an arbitrary subclass of TCP Server as the server to be started
        # Here, it is an Simple HTTP file serving server
        handler = BayernLuftHandler

        max_retries = 10
        r = 0
        while not self.server:
            try:
                # Connect to any open port
                self.server = BayernLuftServer("rev1", (ADDRESS, 0), handler)
            except OSError:
                if r < max_retries:
                    r += 1
                else:
                    raise
                time.sleep(1)

        self.server_control = Server(self.server)
        self.port = self.server_control.get_port()
        self.url = "{}:{}".format(ADDRESS, self.port)
        # Start test server before running any tests
        self.server_control.start_server()

        # Block server to make sure we get "no support"
        self.server.set_blocked()

        try:
            async def fetch():
                async with aiohttp.ClientSession() as session:
                    self.bayernluefter = Bayernluefter(self.url, session)
                    await self.bayernluefter.update()

            loop = asyncio.get_event_loop()
            loop.run_until_complete(fetch())
            self.fail("No error thrown when noticing that the host does not support Syncthru")
        except ValueError:
            pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# coding=utf-8
from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

short_description = '{}'.format(
    'Automated async web-based communication with the Bayernluefter')

setup(
    name='pyernluefter',
    version='0.1.1',
    description=short_description,
    author='nielstron',
    author_email='n.muendler@web.de',
    url='https://github.com/nielstron/pyernluefter/',
    py_modules=['pyernluefter'],
    packages=find_packages(),
    package_data={'pyernluefter.tests.test_structure': ['?export=1', '*.html', '*.txt']},
    install_requires=[
        'aiohttp',
        'parse',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Object Brokering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    keywords='python api iot async bayernluft bayernluefter',
    python_requires='>=3',
    test_suite='pyernluefter.tests',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import threading


class Server:
    def __init__(self, server):
        """
        Expects subclass of TCPServer as argument
        """
        self._server = server
        self._server_started_event = threading.Event()
        self._server_running = False

    def _run_server(self):

        print("Server started, serving on port {}".format(self.get_port()))

        # notify about start
        self._server_started_event.set()
        self._server_running = True

        try:
            self._server.serve_forever()
        finally:
            self._cleanup_server()

    def _cleanup_server(self):
        self._server_running = False
        self._server.server_close()
        # Here, server was stopped
        print("Server stopped")

    def stop_server(self):
        """
        Close server forcibly
        :return:
        """
        print("Stopping server")
        if self._server_running:
            self._server.shutdown()

    def start_server(self, timeout=10):
        """
        Start server thread as daemon
        As such the program will automatically close the thread on exit of all non-daemon threads
        :return:
        """
        self._server_started_event.clear()
        # start webserver as daemon => will automatically be closed when non-daemon threads are closed
        t = threading.Thread(target=self._run_server, daemon=True)
        # Start webserver
        t.start()
        # wait (non-busy) for successful start
        self._server_started_event.wait(timeout=timeout)

    def get_port(self):
        return self._server.server_address[1]



import datetime
from ...convert import SystemMode

RAW_V1 = {
    "Date": "03.05.2020",
    "Time": "20:56:38",
    "DeviceName": "ECFABC0D1111",
    "MAC": "A020A6199C99",
    "LocalIP": "192.168.178.137",
    "RSSI": "-60",
    "FW_MainController": "1838000A",
    "FW_WiFi": "WS181130",
    "SystemMode": "Kellermode",
    "Speed_In": "01",
    "Speed_Out": "01",
    "Speed_AntiFreeze": "00",
    "Temp_In": "0,0",
    "Temp_Out": "18,6",
    "Temp_Fresh": "18,6",
    "rel_Humidity_In": "0,0",
    "rel_Humidity_Out": "77,7",
    "abs_Humidity_In": "0,0",
    "abs_Humidity_Out": "12,4",
    "Efficiency": "N/A",
    "Humidity_Transport": "-1485",
    "_SystemOn": "1",
    "_FrostschutzAktiv": "0",
    "_Frozen": "0",
    "_AbtauMode": "0",
    "_VermieterMode": "0",
    "_QuerlueftungAktiv": "0",
    "_MaxMode": "0",
}

PROCESSED_V1 = {
    "Date": datetime.date(2020, 5, 3),
    "Time": datetime.time(20, 56, 38),
    "DeviceName": "ECFABC0D1111",
    "MAC": "0xa020a6199c99",
    "LocalIP": "192.168.178.137",
    "RSSI": -60,
    "FW_MainController": "1838000A",
    "FW_WiFi": "WS181130",
    "SystemMode": SystemMode.Kellermode,
    "Speed_In": 1,
    "Speed_Out": 1,
    "Speed_AntiFreeze": 0,
    "Temp_In": 0.0,
    "Temp_Out": 18.6,
    "Temp_Fresh": 18.6,
    "rel_Humidity_In": 0.0,
    "rel_Humidity_Out": 77.7,
    "abs_Humidity_In": 0.0,
    "abs_Humidity_Out": 12.4,
    "Efficiency": None,
    "Humidity_Transport": -1485,
    "_SystemOn": True,
    "_FrostschutzAktiv": False,
    "_Frozen": False,
    "_AbtauMode": False,
    "_VermieterMode": False,
    "_QuerlueftungAktiv": False,
    "_MaxMode": False,
}

RAW_V2 = {
    "Date": "11.09.2023",
    "Time": "06:30:36",
    "DeviceName": "3485185xxxxx",
    "MAC": "000000000000",
    "LocalIP": "192.168.0.0",
    "RSSI": "-82",
    "FW_MainController": "Rev2.1",
    "FW_WiFi": "WS32233602",
    "SystemMode": "Zeitgesteuert",
    "_SystemOn": "0",
    "Speed_In": "00",
    "Speed_Out": "00",
    "Speed_AntiFreeze": "00",
    "Temp_In": "0,0",
    "Temp_Out": "0,0",
    "Temp_Fresh": "21,8",
    "rel_Humidity_In": "0,0",
    "rel_Humidity_Out": "0,0",
    "abs_Humidity_In": "0,0",
    "abs_Humidity_Out": "0,0",
    "Efficiency": "N/A",
    "Humidity_Transport": "0",
}

PROCESSED_V2 = {
    "Date": datetime.date(2023, 9, 11),
    "Time": datetime.time(6, 30, 36),
    "DeviceName": "3485185xxxxx",
    "MAC": hex(0x000000000000),
    "LocalIP": "192.168.0.0",
    "RSSI": -82,
    "FW_MainController": "Rev2.1",
    "FW_WiFi": "WS32233602",
    "SystemMode": SystemMode.Zeitgesteuert,
    "_SystemOn": False,
    "Speed_In": 0,
    "Speed_Out": 0,
    "Speed_AntiFreeze": 0,
    "Temp_In": 0.0,
    "Temp_Out": 0.0,
    "Temp_Fresh": 21.8,
    "rel_Humidity_In": 0.0,
    "rel_Humidity_Out": 0.0,
    "abs_Humidity_In": 0.0,
    "abs_Humidity_Out": 0.0,
    "Efficiency": None,
    "Humidity_Transport": 0,
}
